"""
(c) 2020 АО 'НИИ "МАСШТАБ"'
"""

from functools import wraps
import logging
from veilrpc.utils.api.session import ApiResponse

LOG = logging.getLogger("bazalt")


def asyncable(f):
    @wraps(f)
    def wrapped(*args, asynced=False, async_wait=True, async_timeout=None,
                context=None, expected_statuses=(200, 202), **kwargs):
        """Декоратор расширяет функцию,
        добавляя к ней возможность асинхронного выполнения.
        Для совместимости с этим декоратором функция должна
        соответствовать следующим требованиям:
        - принимать в качестве аргумента context;
        - возвращать tuple из (ApiResponse instance, VeilApiObject subclass);
        Если VeilApiObject возвращать не нужно, например, при удалении, можно
         передать (ApiResponse instance, None)
        После того как функция обернута декоратором,
         её сигнатура будет расширена, добавятся аргументы:
          asynced — запросить асинхронное выполнение задачи;
          async_wait — ждать завершения задачи;
          async_timeout — таймаут ожидания завершения задачи;

        Выполнение функции с флагом async вернет Task(VeilApiObject)
        Выполнение функции с флагом async и async_wait вернет Instance объекта
         или None
        При истечении async_timout  raise VigilantApiError
        """
        ctxt = {"async": asynced}
        ctxt.update(context or {})
        r, klass = f(*args, context=ctxt, **kwargs)
        assert isinstance(r, ApiResponse)
        if klass:
            assert issubclass(klass, VeilApiObject)
        r.raise_for_status_code(expected_statuses)
        if r.async_response:
            LOG.debug(f"received async response: {r.data}")
            from veilrpc.utils.api.task import Task, wait_task_ready

            if async_wait:
                task = wait_task_ready(r.task_uuid, timeout=async_timeout)
                task.raise_for_status_success()
                data = task.response
                if 'nodes_user_responses' in task.data and task.data['nodes_user_responses']:
                    LOG.info(f"task response {task.data.nodes_user_responses}")
            else:
                return Task.get(r.task_uuid)
        else:
            LOG.debug(f"received sync response: {r.data}")
            data = r.data
        if klass and data:
            api_obj = klass(data)
            LOG.debug(f"{api_obj.__class__.__name__} {api_obj.uuid} created")
            return api_obj
    return wrapped


class AttributeDict(dict):
    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value


class VeilApiObject:
    api_session = None
    base_url = ""

    class STATUS:
        creating = 'CREATING'
        active = 'ACTIVE'
        failed = 'FAILED'
        deleting = 'DELETING'
        service = 'SERVICE'
        partial = 'PARTIAL'

    @classmethod
    def set_session(cls, session):
        cls.api_session = session

    def __init__(self, data):
        self.data = AttributeDict(data)

    def refresh(self):
        self.data = AttributeDict(self._get(self.uuid))
        LOG.debug(f"{self} data updated")

    @property
    def uuid(self):
        return self.data.id

    @classmethod
    def not_exist(cls, uuid):
        cls._get(uuid, status=404)

    def is_exist(self):
        r = self.api_session.get(self.get_self_url(self.uuid))
        return r.status_code == 200

    @classmethod
    def get(cls, uuid):
        data = cls._get(uuid)
        entity = cls(data)
        LOG.debug(f"{cls.__name__} retrieve")
        return entity

    @classmethod
    def list(cls, **query_params):
        r = cls.api_session.get(cls.base_url, params=query_params)
        r.raise_for_status_code()
        result = []
        for entity in r.list_results:
            result.append(cls(entity))
        LOG.debug(f"{cls.__name__} list")
        return result

    @classmethod
    def list_and_count(cls, **query_params):
        r = cls.api_session.get(cls.base_url, params=query_params)
        r.raise_for_status_code()
        result = []
        for entity in r.list_results:
            result.append(cls(entity))
        LOG.debug(f"{cls.__name__} list")
        return result, r.count_results

    def put(self, **data):
        r = self.api_session.put(self.get_self_url(self.uuid), data=data)
        r.raise_for_status_code()
        return r.data

    @classmethod
    def _get(cls, uuid, status=200):
        r = cls.api_session.get(cls.get_self_url(uuid))
        r.raise_for_status_code(expected_status=status)
        return r.data

    @classmethod
    def get_self_url(cls, uuid=None):
        return f"{cls.base_url}{uuid}/"

    @property
    def url(self):
        return self.get_self_url(self.uuid)

    @staticmethod
    def prepare_url(url, context):
        context = context or dict()
        if context.get("async"):
            if url.endswith("/"):
                url = url + "?async=1"
            else:
                url = url + "&async=1"
        if context.get("dry-run"):
            if url.endswith("/"):
                url = url + "?dry-run=1"
            else:
                url = url + "&dry-run=1"
        return url

    @asyncable
    def simple_instance_action(self, action, data=None, context=None, expected_statuses=(200, 202)):
        LOG.info("{} {} {}".format(self.__class__.__name__, self.name, action))
        r = self.api_session.post(self.prepare_url("{url}{action}/".format(
            url=self.get_self_url(self.uuid),
            action=action), context), data=data)
        return r, None

    def __eq__(self, other):
        result = True

        for field in self.data:
            if getattr(self.data, field) != getattr(other.data, field):
                result = False

        return result

    @property
    def name(self):
        return self.data.name

    @property
    def creating(self):
        return self.data.status == self.STATUS.creating if 'status' in self.data else False

    @property
    def active(self):
        return self.data.status == self.STATUS.active if 'status' in self.data else False

    @property
    def not_active(self):
        return self.data.status != self.STATUS.active if 'status' in self.data else False

    @property
    def failed(self):
        return self.data.status == self.STATUS.failed if 'status' in self.data else False

    @property
    def deleting(self):
        return self.data.status == self.STATUS.deleting if 'status' in self.data else False

    @property
    def service(self):
        return self.data.status == self.STATUS.service if 'status' in self.data else False

    @property
    def built_in(self):
        return self.data.built_in if 'built_in' in self.data else False

    @property
    def partial(self):
        return self.data.status == self.STATUS.partial if 'status' in self.data else False
