"""
(c) 2020 АО 'НИИ "МАСШТАБ"'
"""

import logging
import time

from .base_object import VeilApiObject, asyncable
from functools import partialmethod

LOG = logging.getLogger("bazalt")


class Node(VeilApiObject):
    base_url = "nodes/"

    # pylint: disable=useless-parent-delegation
    def __init__(self, data):
        super().__init__(data)

    @property
    def name(self):
        return self.data.verbose_name

    @classmethod
    @asyncable
    def create(cls, verbose_name, management_ip, controller_ip, ssh_password,
               ssh_user="root", ssh_port=22, description="just node",
               context=None, force=False, datacenter=None):
        data = {
            "management_ip": management_ip,
            "controller_ip": controller_ip,
            "ssh_password": ssh_password,
            "ssh_user": ssh_user,
            "ssh_port": ssh_port,
            "description": description,
            "verbose_name": verbose_name,
            "force": force
        }
        if datacenter is not None:
            data['datacenter'] = datacenter
        r = cls.api_session.post(cls.prepare_url(cls.base_url, context), data=data)
        return r, cls

    def edit(self, verbose_name="", description="", fencing_type=''):
        data = {}
        if verbose_name:
            data.update({"verbose_name": verbose_name})
        if description:
            data.update({"description": description})
        if fencing_type:
            data.update({"fencing_type": fencing_type})
        r = self.api_session.put(self.get_self_url(self.uuid), data=data)
        r.raise_for_status_code()
        LOG.debug("node {} edited".format(self.uuid))

    def lsblk(self, native=False):
        r = self.api_session.get(
            self.get_self_url(self.uuid) + 'lsblk/'
        )
        r.raise_for_status_code()
        if native:
            return r.data
        return [dev.get('name') for dev in r.data]

    def get_volume_groups(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'volume-groups/')
        r.raise_for_status_code()
        return r.list_results

    def get_network_info(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'network-info/')
        r.raise_for_status_code()
        return r.data

    def get_available_volume_groups(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'available-volume-groups/')
        r.raise_for_status_code()
        return r.list_results

    def get_engine_status(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'engine-status/')
        r.raise_for_status_code()
        return r.data

    @asyncable
    def create_volume_group(self, vgname, devices, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'volume-group-create/', context),
                                  data={'vgname': vgname,
                                        'devices': devices})
        r.raise_for_status_code()
        return r, None

    def delete_volume_group(self, vgname):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'volume-group-remove/',
                                  data={'vgname': vgname})
        r.raise_for_status_code()

    def extend_volume_group(self, vgname, devices):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'volume-group-extend/',
                                  data={'vgname': vgname,
                                        'devices': devices})
        r.raise_for_status_code()

    @asyncable
    def restart_minion(self, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'restart-minion/', context),
                                  data={})
        r.raise_for_status_code()
        return r, None

    def iscsi_discovery(self, hostname, port):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'iscsi-discovery/',
                                 params={'hostname': hostname,
                                         'port': port})
        r.raise_for_status_code()
        return r.data

    def fc_discovery(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'fc-discovery/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def discovery_datapool(self):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'discovery-datapool/', params={})
        r.raise_for_status_code()
        return r.data

    def discovery_shared_storage(self):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'discovery-shared-storage/', data={})
        r.raise_for_status_code()
        return r.data

    def discovery_zfs_pool(self):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'discovery-zfs-pool/', params={})
        r.raise_for_status_code()
        return r.data

    def set_status(self, status):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'set-status/',
                                  data={'status': status})
        r.raise_for_status_code()
        return r.data

    set_status_service = partialmethod(set_status, 'SERVICE')
    set_status_active = partialmethod(set_status, 'ACTIVE')

    def ipmi_command(self, command):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'ipmi-command/',
                                  data={'command': command})
        r.raise_for_status_code()
        return r.data

    def fence(self):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'fence/',
                                  data={})
        r.raise_for_status_code()

    def usb_devices(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'usb-devices/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def pci_devices(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'pci-devices/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def pci_tree(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'pci-tree/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def ipmi_ip(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'get-ipmi-ip/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def update_hw_info(self):
        r = self.api_session.post(self.get_self_url(self.uuid) + 'update-hw-info/', data={})
        r.raise_for_status_code()
        LOG.debug("node {} update_hw_info".format(self.uuid))
        return Node(r.data)

    @asyncable
    def migrate_domains(self, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'migrate-domains/', context),
                                  data={})
        r.raise_for_status_code()
        LOG.debug("node {} migrate_domains".format(self.uuid))
        return r, self.__class__

    @asyncable
    def start_domains(self, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'start-domains/', context),
                                  data={})
        r.raise_for_status_code()
        LOG.debug("node {} start_domains".format(self.uuid))
        return r, self.__class__

    @asyncable
    def shutdown_domains(self, timeout=60, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'shutdown-domains/', context),
                                  data={'timeout': timeout})
        r.raise_for_status_code()
        LOG.debug("node {} shutdown_domains".format(self.uuid))
        return r, self.__class__

    @asyncable
    def transfer(self, cluster_id, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'transfer/', context),
                                  data={'cluster': cluster_id})
        r.raise_for_status_code()
        LOG.debug("node {} transfer".format(self.uuid))
        return r, self.__class__

    def save_ipmi_credentials(self, **kwargs):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'save-ipmi-credentials/',
                                 data=kwargs)
        r.raise_for_status_code()
        return r.data

    @classmethod
    def list(cls, **query_params):
        r = cls.api_session.get(cls.base_url, params=query_params)
        r.raise_for_status_code()
        result = []
        for entity in r.list_results:
            result.append(cls(entity))

        LOG.debug("%s list", cls.__name__)
        return result

    def delete(self, ssh_password, force=False):
        if force:
            data = {'force': force}
        else:
            data = {'ssh_password': ssh_password}
        r = self.api_session.post(self.get_self_url(self.uuid) + "remove/",
                                  data=data)
        r.raise_for_status_code()
        LOG.debug("Node deleted\n response data: %s", r.data)

    @property
    def addable_ports(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'addable-ports/')
        r.raise_for_status_code()
        return r.data

    def sync_sriov_ifaces(self):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'sync-sriov-ifaces/')
        r.raise_for_status_code()
        return r.data

    def stats_v2(self, start=time.time() - 59, end=time.time(), step=1):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'stats-v2/?start={}&end={}&step={}'.format(start,
                                                                                                           end,
                                                                                                           step),
                                 params={})
        r.raise_for_status_code()
        return r.data

    def stats_v2_disk_list(self, metric='disk_list'):
        r = self.api_session.get(
            self.get_self_url(self.uuid) + 'stats-v2/?&metric={}'.format(metric),
            params={})
        r.raise_for_status_code()
        return r.data

    def stats_v2_disk(self, start=time.time() - 59, end=time.time(), step=1, metric='disk_iostat', device='vda'):
        r = self.api_session.get(
            self.get_self_url(self.uuid) + 'stats-v2/?start={}&end={}&step={}&metric={}&device={}'.format(start,
                                                                                                          end,
                                                                                                          step,
                                                                                                          metric,
                                                                                                          device),
            params={})
        r.raise_for_status_code()
        return r.data

    def version(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'version/')
        r.raise_for_status_code()
        return r.data

    def domains_resources_used(self):
        # pylint: disable=format-string-without-interpolation
        r = self.api_session.get(self.get_self_url(self.uuid) + 'domains-resources-used/'.format(),
                                 params={})
        r.raise_for_status_code()
        return r.data

    def unknown_domains(self):
        # pylint: disable=format-string-without-interpolation
        r = self.api_session.get(self.get_self_url(self.uuid) + 'unknown-domains/'.format(),
                                 params={})
        r.raise_for_status_code()
        return r.data

    def unknown_domains_count(self):
        # pylint: disable=format-string-without-interpolation
        r = self.api_session.get(self.get_self_url(self.uuid) + 'unknown-domains-count/'.format(),
                                 params={})
        r.raise_for_status_code()
        return r.data

    @asyncable
    def delete_unknown_domain(self, domain_names, delete_all=True, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'delete-unknown-domain/', context),
                                  data={'domain_names': domain_names,
                                        'delete_all': delete_all})
        r.raise_for_status_code()
        return r, self.__class__

    @property
    def cpu_count(self):
        if 'cpu_topology' not in self.data:
            self.refresh()
        return self.data.cpu_topology['cpu_count']

    @property
    def cpu_map(self):
        if 'cpu_topology' not in self.data:
            self.refresh()
        return self.data.cpu_topology['cpu_map']

    @property
    def cpu_type(self):
        if 'cpu_topology' not in self.data:
            self.refresh()
        return self.data.cpu_topology['cpu_type']

    @property
    def cpu_features(self):
        if 'cpu_topology' not in self.data:
            self.refresh()
        return self.data.cpu_topology['cpu_features']

    @asyncable
    def change_ip(self, data, expected_status=(200, 202), context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'change-node-ip/', context),
                                  data=data)
        r.raise_for_status_code(expected_status=expected_status)
        return r, None

    def drs_stats(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'drs-stats/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def usage(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'usage/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def get_ksm_stat(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'get-ksm-stat/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def ksm(self, data, expected_status=(200, 202)):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'ksm/', data=data)
        r.raise_for_status_code(expected_status=expected_status)

    def get_swappiness(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'swappiness/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    @asyncable
    def update_swappiness(self, swappiness=None, vfs_cache_pressure=None, expected_status=(200, 202), context=None):
        data = {}
        if swappiness:
            data['swappiness'] = swappiness
        if vfs_cache_pressure:
            data['vfs_cache_pressure'] = vfs_cache_pressure
        r = self.api_session.put(self.prepare_url(self.get_self_url(self.uuid) + 'swappiness/', context),
                                 data=data)
        r.raise_for_status_code(expected_status=expected_status)
        return r, self.__class__

    def get_iommu(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'iommu/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    @asyncable
    def update_iommu(self, iommu, expected_status=(200, 202), context=None):
        r = self.api_session.put(self.prepare_url(self.get_self_url(self.uuid) + 'iommu/', context),
                                 data={"iommu": iommu})
        r.raise_for_status_code(expected_status=expected_status)
        return r, self.__class__

    def save_fence_settings(self, fencing_type=None, heartbeat_type=None, datapools=None, expected_status=(200, 202)):
        data = {}
        if fencing_type:
            data['fencing_type'] = fencing_type
        if heartbeat_type:
            data['heartbeat_type'] = heartbeat_type
        if datapools:
            data['datapools'] = datapools
        r = self.api_session.put(self.get_self_url(self.uuid) + 'save-fence-settings/',
                                 data=data)
        r.raise_for_status_code(expected_status=expected_status)

    def upgrade_check(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'upgrade-check/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    @asyncable
    def upgrade_start(self, expected_status=(200, 202), context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'upgrade-start/', context),
                                  data={})
        r.raise_for_status_code(expected_status=expected_status)
        return r, None

    def smart(self, brief=None):
        r = self.api_session.get(self.get_self_url(self.uuid) +
                                 'smart/{}'.format('' if brief is None else '?brief={}'.format(brief)))
        r.raise_for_status_code()
        return r.data

    def ballooning(self, expected_status=(200, 202), context=True):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'ballooning/', data={"ballooning": context})
        r.raise_for_status_code(expected_status=expected_status)

    def quorum_info(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'quorum-info/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def get_available_mdev(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'list-available-mdev/', params={})
        r.raise_for_status_code()
        return r.data

    def create_mdev(self, pci_id, md_type, expected_status=(200,), context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'create-mdev/', context),
                                  data={'md_type': md_type, 'pci_id': pci_id})
        r.raise_for_status_code(expected_status=expected_status)
        return r.data

    def remove_mdev(self, uuid, expected_status=(200,), context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'remove-mdev/', context),
                                  data={'mdev': uuid})
        r.raise_for_status_code(expected_status=expected_status)
        return r.data

    def available_cpu_models(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'available-cpu-models/', params={})
        r.raise_for_status_code()
        return r.data

    @classmethod
    def test_connection(cls, hostname, node=None, ssh_password=None, ssh_user=None,
                        ssh_port=None):
        url = cls.base_url + f'test-connection/?hostname={hostname}'
        if node is not None:
            url += f'&node={node}'
        if ssh_password is not None:
            url += f'&ssh_password={ssh_password}'
        if ssh_user is not None:
            url += f'&ssh_user={ssh_user}'
        if ssh_port is not None:
            url += f'&ssh_port={ssh_port}'
        r = cls.api_session.get(url)
        r.raise_for_status_code()
        return r.data

    @property
    def render_devices(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'render-devices/', params={})
        r.raise_for_status_code()
        return r.data

    def connected_wwns(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'connected-wwns/', params={})
        r.raise_for_status_code()
        return r.data

    def local_wwns(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'local-wwns/', params={})
        r.raise_for_status_code()
        return r.data

    def iommu_state(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'iommu-state/', params={})
        r.raise_for_status_code()
        return r.data
